#! /usr/bin/env python

# import standard modules
from optparse import OptionParser
import os, re, sys, time

# import custom modules
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import SetupMainClass, TimeSpent, WriteCommand
from utils.messages import ErrMsg, LogMsg, DebugMsg
from utils.files_paths import (FileBoxCls, CheckFilePath, CheckDirPath,
                               GetOutDir, EnsureAbsPath)
from parsers.candidate_group_parser import CandidateGroupParserCls

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3

class ExonBoundCounterCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
    self.exon_bound_coords = dict()
    self.output_file = None
  #} end def

  def __del__(self): #{
    CloseLogFile(self)
    if (None != self.output_file): #{
      self.output_file.Close()
    #} end if
  #} end def

  def CountExonBoundaryJunctions(self): #{
    LogMsg(self, "Counting exon-boundary matching junction coordinates...")
    start = time.time()
    # load the exon boundary coordinates from the annotations file(s)
    self.LoadExonBoundaryCoordinates()
    # create the parser
    parser = CandidateGroupParserCls(self.options.input_file)
    # create the output file
    self.CreateOutputFile()
    # iterate through the events
    for event in parser: #{
      # count member junction coordinates matching exon boundaries
      self.ProcessEvent(event)
    #} end for
    LogMsg(self, "Time spent counting exon-boundary matching "
      "junction coordinates: %s" % TimeSpent(start))
  #} end def

  def LoadExonBoundaryCoordinates(self): #{
    LogMsg(self, "Loading exon boundary coordinates...")
    start = time.time()
    # load the exon boundary coordinates from the main annotations file
    self.LoadExonBoundaryCoordinatesFromFile(self.options.exon_coords_path)
    # load exon boundary coordinates from additional annotations files?
    LogMsg(self, "Time spent loading exon boundary coordinates: %s" %
      TimeSpent(start))
  #} end def

  def LoadExonBoundaryCoordinatesFromFile(self, path): #{
    DebugMsg(self, "Exon file: %s" % path)
    # open the annotations file
    fail_msg = "cannot read exon boundary coordinates file"
    exons_file = FileBoxCls(path, "r", fail_msg)
    # get the coordinates from the file
    for exon_line in exons_file: #{
      (chrom, left, right, name) = exon_line.split(" ", 3)
      # skip UTRs and non_coding genes
      if (name.lower().endswith("utr") or
          name.lower().endswith("non_coding")):
        continue
      #} end if
      left = int(left)
      right = int(right)
      # fix chromosome names, if needed
      chrom = chrom.replace("chr", "").replace("MT", "M")
      chrom_patt = r"^(\d+|[XYM])$"
      if (None != re.search(chrom_patt, chrom)): #{
        if (chrom in self.exon_bound_coords): #{
          left_coords = self.exon_bound_coords[chrom]['left']
          right_coords = self.exon_bound_coords[chrom]['right']
          if (left in left_coords): #{
            if (right in left_coords[left]): #{
              left_coords[left][right].append(name)
            else:
              left_coords[left][right] = list([name])
            #} end if
          else:
            left_coords[left] = dict({right: list([name])})
          #} end if
          if (right in right_coords): #{
            if (left in right_coords[right]): #{
              right_coords[right][left].append(name)
            else:
              right_coords[right][left] = list([name])
            #} end if
          else:
            right_coords[right] = dict({left: list([name])})
          #} end if
        else:
          self.exon_bound_coords[chrom] = dict({
            'left':  dict({left:  dict({right: list([name])})}),
            'right': dict({right: dict({left:  list([name])})}),
          })
        #} end if
      else:
        DebugMsg(self, "Skipping exon in strange chromosome: %s" % exon_line)
      #} end if
    #} end for
    # close the file
    exons_file.close()
  #} end def

  def CreateOutputFile(self): #{
    input_file_name = os.path.basename(self.options.input_file)
    output_path = os.path.join(self.options.output_dir, input_file_name)
    fail_msg = "could not create exon-boundary output file"
    self.output_file = FileBoxCls(output_path, "w", fail_msg)
  #} end def

  def ProcessEvent(self, event): #{
    DebugMsg(self, "Processing event %i..." % event.id)
    event.exon_bounds = 0
    # iterate through the event's members
    for member in event.members: #{
      self.ProcessMember(member)
      if (member.meta_fields['exon_bounds'] > event.exon_bounds): #{
        event.exon_bounds = member.meta_fields['exon_bounds']
      #} end if
    #} end for
    # write the output
    self.output_file.Write(event.FullDataString())
  #} end def

  def ProcessMember(self, member): #{
    DebugMsg(self, "Processing member %s..." % member.IDString())
    member.meta_fields['exon_bounds'] = 0
    # if the member is an internal gap
    if (member.gap and member.GapIsInternal()): #{
      # test the left (downstream) side of the gap
      junction_coord = min(member.align_info_B.genome_start,
        member.align_info_B.genome_end)
      if (self.JunctionMatchesBoundary
          (member.align_info_B.chrom, 'left', junction_coord)):
        member.meta_fields['exon_bounds'] += 1
      #} end if
      # test the right (upstream) side of the gap
      junction_coord = max(member.align_info_B.genome_start,
        member.align_info_B.genome_end)
      if (self.JunctionMatchesBoundary
          (member.align_info_B.chrom, 'right', junction_coord)):
        member.meta_fields['exon_bounds'] += 1
      #} end if
    else:
      # if treating an edge gap event as a split event,
      # alignments might need to be reordered to be in contig order
      if (member.gap and
          member.align_info_A.ctg_start > member.align_info_B.ctg_start):
        DebugMsg(self, "Reordering alignments in gapped split event...")
        align_info_A = member.align_info_B
        align_info_B = member.align_info_A
      else:
        align_info_A = member.align_info_A
        align_info_B = member.align_info_B
      #} end if
      # test the first junction coordinate
      if ("+" == align_info_A.Strand()): #{
        junction_side = 'right'
      else:
        junction_side = 'left'
      #} end if
      junction_coord = align_info_A.genome_end
      if (self.JunctionMatchesBoundary
          (align_info_A.chrom, junction_side, junction_coord)):
        member.meta_fields['exon_bounds'] += 1
      #} end if
      # test the second junction coordinate
      if ("+" == align_info_B.Strand()): #{
        junction_side = 'left'
      else:
        junction_side = 'right'
      #} end if
      junction_coord = align_info_B.genome_start
      if (self.JunctionMatchesBoundary
          (align_info_B.chrom, junction_side, junction_coord)):
        member.meta_fields['exon_bounds'] += 1
      #} end if
    #} end if
  #} end def

  def JunctionMatchesBoundary(self, chrom, j_side, j_coord): #{
    if (chrom not in self.exon_bound_coords): #{
      LogMsg(self, "Warning: no exons on chromosome %s" % chrom)
      return False
    #} end if
    DebugMsg(self, "Junction side: %s Junction coord: %s" % (j_side, j_coord))
    offset_range = range(-self.options.buffer_size, self.options.buffer_size+1)
    for offset in offset_range: #{
      test_coord = j_coord + offset
      DebugMsg(self, "Offset: %s Test coord: %s" % (offset, test_coord))
      if (test_coord in self.exon_bound_coords[chrom][j_side]): #{
        return True
      #} end if
    #} end if
    return False
  #} end def
#} end class

#### EXCEPTION CLASSES ####
class ExonBoundCounterError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = ("Go through the group members in the input file and "
    "for each member check how many of its junction coordinates match up "
    "with exon boundaries.")
  args = [ "GROUPS_FILE", "EXON_COORDINATES", ]
  usage_string = ("%prog " + " ".join(args) + " [ OPTIONS ]\n\n"
    "The exon coordinates file should be space-delimited with the following "
    "columns: \"chr start end id\"")
  parser = OptionParser(description=description_string,
                        version="%prog 0.1.2",
                        usage=usage_string)
  parser.num_args = len(args)
  parser.add_option("-b", "--buffer-size",
                    type="int", metavar="N",
                    help="Count the junction coordinate as matching an exon "
                         "coordinate if it is within Nbp to either side of "
                         "it. [default:%default]")
  # add option for additional annotations files?
  parser.add_option("-d", "--debug",
                    action="store_true",
                    help="Print debug information while the program runs.")
  parser.set_defaults(buffer_size=4,
                      force=False,
                      debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  path_errors = list()
  CheckFilePath(options.input_file, "BARNACLE results", path_errors)
  input_dir = os.path.dirname(options.input_file)
  CheckFilePath(options.exon_coords_path, "exon coordinates", path_errors)
  options.output_dir = GetOutDir(input_dir, "with_exon_bounds")
  CheckDirPath(options.output_dir, "output", path_errors, create=True)
  # set the log-file name
  options.log_file_name = GetLogPath(options.input_file,
    "exon_bounds", options.output_dir)
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  # the paths are good if there are no path errors
  return (0 == len(path_errors))
#} end def

def Main(): #{
  # get options and arguments
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.input_file  = EnsureAbsPath(args[0])
    options.exon_coords_path = EnsureAbsPath(args[1])
    if (CheckPaths(options)): #{
      try:
        exon_bound_counter = ExonBoundCounterCls(options)
        WriteCommand(exon_bound_counter, sys.argv)
        exon_bound_counter.CountExonBoundaryJunctions()
      except (MyError), e:
        ErrMsg("ERROR while counting exon boundary matching junction "
               "coordinates for BARNACLE event members:\n  %s" % e)
        return ES_RUN_ERR
      # end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify the path to a file containing BARNACLE "
                 "group data (GROUPS_FILE) and the path to a file containing "
                 "exon boundary coordinates (EXON_COORDINATES).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  exit_status = Main()
  if (ES_SUCCESS == exit_status): #{
    print "EXON BOUNDARY COUNTING SUCCESS"
  else:
    print "EXON BOUNDARY COUNTING FAIL"
  #} end if
  sys.exit(exit_status)
#} end if
