#! /usr/bin/env python

# import standard modules

# import custom modules
from utils.general import SetupMainClass
from utils.messages import LogMsg, DebugMsg
from utils.files_paths import GetFilePath, FileBoxCls
from parsers.candidate_group import CandidateGroupCls
from parsers.grouped_candidate import GroupedCandidateCls
from parsers.candidate_group_parser import CandidateGroupParserCls

class BasePredictorCls:
  def __init__(self, key, ext, description, options, log_info=dict()):
    SetupMainClass(self, options, log_info=log_info)
    self.key = key
    self.ext = ext
    self.description = description
    output_dir = options.output_dir
    self.write_pretty = True
    self.write_special = True
    if (options.realign): #{
      output_dir = options.realign_dir
      self.write_pretty = False
      self.write_special = False
    #} end if
    output_path = GetFilePath(output_dir, options.lib, self.ext)
    self.output_file = FileBoxCls(output_path, "w",
      "cannot create %s output file" % self.description)
    if (options.pretty): #{
      pretty_path = GetFilePath(options.output_dir, options.lib,
        "%s.pretty" % self.ext)
      self.pretty_file = FileBoxCls(pretty_path, "w",
        "cannot create pretty %s output file" % self.description)
    else:
      self.pretty_file = None
    #} end if
    self.num_predictions = 0
    # only certain event types require the sequence for the realignment test
    #self.store_seq = False
  # end def

  def __del__(self): #{
    self.output_file.Close()
  #} end def

  def ProcessGroup(self, group, good_members=list()): #{
    DebugMsg(self, "Testing whether group is a %s:\n%s" %
      (self.description.upper(), group.DataString()))
    # check whether group satisfies event-type criteria
    if (self.TestGroup(group)): #{
      event = CandidateGroupCls(group.DataString())
    else:
      DebugMsg(self, "Group is not a %s" % self.description)
      return False
    #} end if
    for (index, member) in enumerate(group.members): #{
      DebugMsg(self, "  Testing whether candidate contig is a %s:\n  %s" %
        (self.description, member.DataString()))
      # check whether member satisfies event-type criteria
      if (self.TestMember(member)): #{
        event.members.append(GroupedCandidateCls(member.DataString()))
        good_members.append(index)
      else:
        DebugMsg(self, "  Member is not a %s" % self.description)
        continue
      #} end if
    #} end for
    # if the no candidate contigs satisfy the event-type criteria
    event.UpdateMemberCount()
    if (0 == event.num_members): #{
      DebugMsg(self, "Group has no %s members" % self.description)
      return False
    #} end if
    DebugMsg(self, "Predicting %s event from group!" % self.description)
    event.event_type = self.key
    self.OutputEvent(event)
    return True
  #} end def

  def OutputEvent(self, event): #{
    self.num_predictions += 1
    # write the event to the output file(s)
    self.output_file.Write(event.FullDataString())
    if (None != self.pretty_file and self.write_pretty): #{
      self.pretty_file.Write(event.FullDataString(readable=True))
    #} end if
  #} end def

  #def ReprocessPredictions(self, contig_alignments, contig_sequences): #{
  def ReprocessPredictions(self, realigned_contigs): #{
    LogMsg(self, "Reprocessing %s predictions..." % self.description)
    self.num_predictions = 0
    self.write_pretty = True
    self.write_special = True
    self.output_file.Close()
    parser = CandidateGroupParserCls(self.output_file.path)
    output_path = GetFilePath(self.options.output_dir, self.options.lib,
      self.ext)
    self.output_file = FileBoxCls(output_path, "w",
      "cannot create %s output file" % self.description)
    for group in parser: #{
      DebugMsg(self, "Reprocessing group %i:" % group.id)
      event = CandidateGroupCls(group.DataString())
      for (index, member) in enumerate(group.members): #{
        DebugMsg(self, "%s %s %s(%ibp):%.2f %s(%ibp):%.2f %s" % (
          member.IDString(), member.contig_info.ToString(with_kform=False),
          member.align_info_A.ContigCoordsString(),
          member.align_info_A.ContigSpan(), member.align_info_A.identity,
          member.align_info_B.ContigCoordsString(),
          member.align_info_B.ContigSpan(), member.align_info_B.identity,
          member.MainGenesString()))
        realigned_contig = realigned_contigs[member.contig_info.id]
        #contig_seq = None
        #if (self.store_seq): #{
        #  contig_seq = contig_sequences[member.contig_info.id]
        #} end if
        #if (self.ReprocessMember(member, contig_alignment, contig_seq)): #{
        if (None == realigned_contig.best_align): #{
          LogMsg(self, "  No alignments for contig %s" % member.contig_info.id)
          continue
        #} end if
        if (self.ReprocessMember(member, realigned_contig)): #{
          event.members.append(group.members[index])
        #} end if
      #} end for
      if (0 == len(event.members)): #{
        DebugMsg(self, "Group %i has no good members after realignment" %
          group.id)
        continue
      #} end if
      event.UpdateMemberCount()
      self.OutputEvent(event)
    #} end for
  #} end def
# end class
